/**
 * DevExtreme (ui/scheduler/workspaces/ui.scheduler.virtual_scrolling.js)
 * Version: 20.2.0 (build 20212-0316)
 * Build date: Thu Jul 30 2020
 *
 * Copyright (c) 2012 - 2020 Developer Express Inc. ALL RIGHTS RESERVED
 * Read about DevExtreme licensing here: https://js.devexpress.com/Licensing/
 */
"use strict";
exports.default = void 0;
var _type = _interopRequireDefault(require("../../../core/utils/type"));

function _interopRequireDefault(obj) {
    return obj && obj.__esModule ? obj : {
        "default": obj
    }
}

function _typeof(obj) {
    "@babel/helpers - typeof";
    if ("function" === typeof Symbol && "symbol" === typeof Symbol.iterator) {
        _typeof = function(obj) {
            return typeof obj
        }
    } else {
        _typeof = function(obj) {
            return obj && "function" === typeof Symbol && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj
        }
    }
    return _typeof(obj)
}

function _toConsumableArray(arr) {
    return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _unsupportedIterableToArray(arr) || _nonIterableSpread()
}

function _nonIterableSpread() {
    throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")
}

function _unsupportedIterableToArray(o, minLen) {
    if (!o) {
        return
    }
    if ("string" === typeof o) {
        return _arrayLikeToArray(o, minLen)
    }
    var n = Object.prototype.toString.call(o).slice(8, -1);
    if ("Object" === n && o.constructor) {
        n = o.constructor.name
    }
    if ("Map" === n || "Set" === n) {
        return Array.from(o)
    }
    if ("Arguments" === n || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) {
        return _arrayLikeToArray(o, minLen)
    }
}

function _iterableToArray(iter) {
    if ("undefined" !== typeof Symbol && Symbol.iterator in Object(iter)) {
        return Array.from(iter)
    }
}

function _arrayWithoutHoles(arr) {
    if (Array.isArray(arr)) {
        return _arrayLikeToArray(arr)
    }
}

function _arrayLikeToArray(arr, len) {
    if (null == len || len > arr.length) {
        len = arr.length
    }
    for (var i = 0, arr2 = new Array(len); i < len; i++) {
        arr2[i] = arr[i]
    }
    return arr2
}

function _classCallCheck(instance, Constructor) {
    if (!(instance instanceof Constructor)) {
        throw new TypeError("Cannot call a class as a function")
    }
}

function _defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
        var descriptor = props[i];
        descriptor.enumerable = descriptor.enumerable || false;
        descriptor.configurable = true;
        if ("value" in descriptor) {
            descriptor.writable = true
        }
        Object.defineProperty(target, descriptor.key, descriptor)
    }
}

function _createClass(Constructor, protoProps, staticProps) {
    if (protoProps) {
        _defineProperties(Constructor.prototype, protoProps)
    }
    if (staticProps) {
        _defineProperties(Constructor, staticProps)
    }
    return Constructor
}
var ROW_HEIGHT = 50;
var VirtualScrolling = function() {
    function VirtualScrolling(workspace, viewportHeight, scrollable) {
        _classCallCheck(this, VirtualScrolling);
        this._workspace = workspace;
        this._viewportHeight = viewportHeight;
        this._scrollable = scrollable;
        this._init()
    }
    _createClass(VirtualScrolling, [{
        key: "getWorkspace",
        value: function() {
            return this._workspace
        }
    }, {
        key: "getViewportHeight",
        value: function() {
            return this._viewportHeight
        }
    }, {
        key: "getScrollable",
        value: function() {
            return this._scrollable
        }
    }, {
        key: "getRowHeight",
        value: function() {
            return ROW_HEIGHT
        }
    }, {
        key: "getState",
        value: function() {
            return this._state
        }
    }, {
        key: "getLayoutMap",
        value: function() {
            return this.getState().layoutMap
        }
    }, {
        key: "_setLayoutMap",
        value: function(layoutMap) {
            return this._state.layoutMap = layoutMap
        }
    }, {
        key: "_init",
        value: function() {
            var pageSize = Math.ceil(this.getViewportHeight() / this.getRowHeight());
            var scrollOffset = {
                top: 0,
                left: 0
            };
            this._state = {
                pageSize: pageSize,
                scrollOffset: scrollOffset,
                startIndex: -1,
                rowCount: 0,
                topVirtualRowCount: 0,
                bottomVirtualRowCount: 0,
                layoutMap: {
                    topVirtualRowHeight: 0,
                    dataItems: [],
                    bottomVirtualRowHeight: 0
                }
            };
            this._updateState(scrollOffset);
            this._attachScrollableEvent()
        }
    }, {
        key: "_attachScrollableEvent",
        value: function() {
            var _this = this;
            var scrollable = this.getScrollable();
            var onScroll = scrollable.option("onScroll");
            scrollable.option("onScroll", function(e) {
                if (onScroll) {
                    onScroll.apply(scrollable, [e])
                }
                var scrollOffset = null === e || void 0 === e ? void 0 : e.scrollOffset;
                if (scrollOffset && _this._updateState(scrollOffset)) {}
            })
        }
    }, {
        key: "_updateState",
        value: function(scrollOffset) {
            var workspace = this.getWorkspace();
            var state = this.getState();
            var top = scrollOffset.top;
            var currentStartIndex = state.startIndex;
            var rowHeight = this.getRowHeight();
            state.scrollOffset = scrollOffset;
            state.startIndex = Math.floor(top / rowHeight);
            if (currentStartIndex !== state.startIndex) {
                var pageSize = state.pageSize;
                var groupCount = workspace._getGroupCount();
                var totalRowCount = workspace._getTotalRowCount(groupCount);
                var topVirtualRowCount = Math.floor(top / rowHeight);
                var deltaRowCount = totalRowCount - topVirtualRowCount;
                var rowCount = deltaRowCount >= pageSize ? pageSize : deltaRowCount;
                var bottomVirtualRowCount = totalRowCount - topVirtualRowCount - rowCount;
                state.topVirtualRowCount = topVirtualRowCount;
                state.rowCount = rowCount;
                state.bottomVirtualRowCount = bottomVirtualRowCount;
                this._updateLayoutMap();
                this._validateState();
                return true
            }
            return false
        }
    }, {
        key: "_updateLayoutMap",
        value: function() {
            var workspace = this.getWorkspace();
            var groupCount = workspace._getGroupCount();
            var options = {
                cellCount: workspace._getTotalCellCount(groupCount),
                cellClass: workspace._getDateTableCellClass.bind(workspace),
                rowClass: workspace._getDateTableRowClass(),
                cellTemplate: workspace.option("dataCellTemplate"),
                getCellData: workspace._getCellData.bind(workspace),
                allDayElements: workspace._insertAllDayRowsIntoDateTable() ? workspace._allDayPanels : void 0,
                groupCount: groupCount,
                groupByDate: workspace.option("groupByDate")
            };
            this._updateLayoutMapCore(options)
        }
    }, {
        key: "_updateLayoutMapCore",
        value: function(options) {
            var state = this.getState();
            var rowHeight = this.getRowHeight();
            var topVirtualRowCount = state.topVirtualRowCount;
            var bottomVirtualRowCount = state.bottomVirtualRowCount;
            var dataRowCount = state.rowCount;
            var rowCountInGroup = dataRowCount;
            if (options.groupCount) {
                rowCountInGroup = dataRowCount / options.groupCount
            }
            var layoutMap = this.getLayoutMap();
            var dataItems = layoutMap.dataItems.slice(0, dataRowCount);
            var prevTopVirtualRowHeight = layoutMap.topVirtualRowHeight;
            var prevBottomVirtualRowHeight = layoutMap.bottomVirtualRowHeight;
            var topVirtualRowHeight = rowHeight * topVirtualRowCount;
            var bottomVirtualRowHeight = rowHeight * bottomVirtualRowCount;
            var addRowCount = 0;
            var isAppend = prevTopVirtualRowHeight < topVirtualRowHeight;
            var isPrepend = prevBottomVirtualRowHeight < bottomVirtualRowHeight;
            var needAddRows = isAppend || isPrepend;
            if (needAddRows) {
                if (isAppend) {
                    addRowCount = Math.ceil((topVirtualRowHeight - prevTopVirtualRowHeight) / rowHeight)
                } else {
                    addRowCount = Math.ceil((bottomVirtualRowHeight - prevBottomVirtualRowHeight) / rowHeight)
                }
                addRowCount = Math.min(addRowCount, dataRowCount);
                dataItems = this._populateLayoutMap(options, dataItems, addRowCount, rowCountInGroup, isAppend);
                layoutMap.dataItems = dataItems;
                layoutMap.topVirtualRowHeight = topVirtualRowHeight;
                layoutMap.bottomVirtualRowHeight = bottomVirtualRowHeight
            }
        }
    }, {
        key: "_populateLayoutMap",
        value: function(options, dataItems, addRowCount, rowCountInGroup, isAppend) {
            var state = this.getState();
            var dataRowCount = state.rowCount;
            var topVirtualRowCount = state.topVirtualRowCount;
            var isEmptyData = 0 === dataItems.length;
            if (!isEmptyData) {
                var deltaRowCount = 0;
                if (dataRowCount > dataItems.length) {
                    deltaRowCount = dataRowCount - dataItems.length
                }
                if (isAppend) {
                    dataItems = dataItems.slice(addRowCount)
                } else {
                    dataItems.splice(-addRowCount)
                }
                addRowCount += deltaRowCount
            }
            if (isAppend) {
                for (var index = 0; index < addRowCount; ++index) {
                    var rowIndex = dataItems.length + topVirtualRowCount;
                    var rowModel = this._createLayoutMapRow(options, rowIndex, rowCountInGroup);
                    dataItems.push(rowModel)
                }
            } else {
                var startIndex = state.startIndex;
                var prependRows = new Array(addRowCount);
                for (var _index = addRowCount - 1; _index >= 0; --_index) {
                    var _rowIndex = startIndex + _index;
                    var _rowModel = this._createLayoutMapRow(options, _rowIndex, rowCountInGroup);
                    prependRows[_index] = _rowModel
                }
                dataItems = [].concat(prependRows, _toConsumableArray(dataItems))
            }
            return dataItems
        }
    }, {
        key: "_createLayoutMapRow",
        value: function(options, rowIndex, rowCountInGroup) {
            var cells = this._populateLayoutMapCells(options, rowIndex);
            var isLastRowInGroup = (rowIndex + 1) % rowCountInGroup === 0;
            var rowModel = {
                rowIndex: rowIndex,
                className: options.rowClass,
                cells: cells,
                needInsertAllDayRow: options.allDayElements && isLastRowInGroup
            };
            return rowModel
        }
    }, {
        key: "_populateLayoutMapCells",
        value: function(options, rowIndex) {
            var cellsMap = [];
            var cellClass = options.cellClass;
            for (var cellIndex = 0; cellIndex < options.cellCount; cellIndex++) {
                var cellModel = {
                    cellIndex: cellIndex
                };
                if (cellClass) {
                    if (options.cellClass) {
                        if (_type.default.isFunction(options.cellClass)) {
                            cellModel.className = options.cellClass(rowIndex, cellIndex)
                        } else {
                            cellModel.className = options.cellClass
                        }
                    }
                }
                cellModel.cellTemplate = options.cellTemplate;
                cellsMap.push(cellModel)
            }
            return cellsMap
        }
    }, {
        key: "dispose",
        value: function() {}
    }, {
        key: "_getStateLogRecursive",
        value: function(target) {
            var _this2 = this;
            var propertyNames = Object.getOwnPropertyNames(target);
            var logArgs = [];
            propertyNames.forEach(function(name) {
                var property = target[name];
                if (!property || "object" !== _typeof(property)) {
                    logArgs.push("".concat(name, ": ").concat(property))
                } else {
                    logArgs.push("\n\t".concat(name, ":"));
                    var args = _this2._getStateLogRecursive(property);
                    logArgs.push.apply(logArgs, _toConsumableArray(args))
                }
            });
            return logArgs
        }
    }, {
        key: "_validateState",
        value: function() {
            var dataItems = this.getLayoutMap().dataItems;
            for (var i = 1; i < dataItems.length; ++i) {
                var rowIndex = dataItems[i].rowIndex;
                var prevRowIndex = dataItems[i - 1].rowIndex;
                var diffRowIndex = rowIndex - prevRowIndex;
                if (1 !== diffRowIndex) {
                    var delta = Math.abs(rowIndex - prevRowIndex);
                    throw "Error Row indices: rowIndex=".concat(rowIndex, ", prevIndex=").concat(prevRowIndex, ", delta=").concat(delta)
                }
            }
            return true
        }
    }]);
    return VirtualScrolling
}();
exports.default = VirtualScrolling;
module.exports = exports.default;
